#!/bin/sh
#
# Component: qr
#
# Description:
#     Starts retrieving/moving study from a query retrieve device
#
# usage:
#  retrieve.sh [-u user] -h <hostname> {-d <destinationhost> | -a <destination AE>} [-f <dumpfile> | -s <study>] [-w wait-timeout]
#
# -f <dumpfile> is only for backward compatibilty (stuck tasks)
#
# -d <destinationhost>  is the ID of the device the study to be moved (SELF means the local server)
# -a <destination AE> the AE title of the device the study to be moved.
# -w <wait-timeout> the wait-timeout in seconds, default is 30
#
# If both destination host and destination AE are specified, destination AE title will be used. 
#
# Since DICOM C-MOVE uses AE title to indentify the destination device. AE Titles should be unique within a facility.
#
#

. $HOME/etc/setenv.sh


# if retrieve.sh is called directly:
# run as a task instead
if [ `basename $0` = "retrieve.sh" ] ; then
	$TASKD_HOME/runshellcmd.sh -R -r 0 $0 "$@"
	exit $?
fi


LOGGER="/usr/bin/logger"
TAG="qrretrieve"

if [ -z "$RETRY_NUM" ] ; then
	RETRY_NUM=0
	export RETRY_NUM
fi

file=""
host=""
destDev=""
destAE=""
study=""
user=""
removeDmpFile="false"
waitTimeout=""

INV_PARAM=""

while getopts 'a:h:d:f:s:u:' OPT; do
	case $OPT in
	h) host="$OPTARG" ;;
	a) destAE="$OPTARG" ;;
	d) destDev="$OPTARG" ;;
	f) file="$OPTARG" ;;
	s) study="$OPTARG" ;;
	u) user="$OPTARG" ;;
	*) INV_PARAM="Invalid_Parameter:$OPT" ;;
	esac
done

logDest="$destAE"
[ -z "$logDest" ] && logDest="$destDev"
[ -z "$logDest" ] && logDest="-"


# Removes refcount in case of error
remove_refcount_()
{
	if [ "$destDev" = "SELF" ] ; then
		fullHost="${host}"
		[ -f "$CONF/serverid.cfg" ] && . $CONF/serverid.cfg
		[ -n "$SELF_SERVER_ID" ] && fullHost="${host}@${SELF_SERVER_ID}"

		$TASKD_HOME/createchildtask.sh -p "QR_CLEAR_ARCHIVE_REF" -t "class" -i "2" -c "qr.ClearArchiveRef -s \"$study\" -r \"archive_${fullHost}_get\""
	fi
}

# create log entries
#
log_()
{
	logUser="$user"
	if [ -z "$1" ] ; then
		$LOGGER -p $LOG_OPER_FACILITY.info -t "$TAG" "INFO STUDY qr_get $logUser - $study - - $host - - - - - - Retrieve_to_$logDest - - - - -"
		$LOG_HOME/logintodatabase.sh -p "INFO" -l "STUDY" -a "qr_get" -u "$logUser" -s "$study" -h "$host" -m "Retrieve_to_$logDest" 
	else
		logMessage=`sed 's; ;_;g' <<<"$1"`
		$LOGGER -p $LOG_OPER_FACILITY.err -t "$TAG" "ERROR STUDY qr_get $logUser - $study - - $host - - - - - - $logMessage - - - - -"
		$LOG_HOME/logintodatabase.sh -p "ERROR" -l "STUDY" -a "qr_get" -u "$logUser" -s "$study" -h "$host" -m "$logMessage" 
	fi
}


#
# Override default run_ method
#
run_()
{
	$LOGGER -p $LOG_FACILITY.debug -t "$TAG" "Start retrieving $study from $host to $logDest"
	
	if [ -n "$INV_PARAM" ] ; then
		echo "$INV_PARAM"
		return 125
	fi

	SELF_SERVER_ID=""
	[ -f "$CONF/serverid.cfg" ] && . $CONF/serverid.cfg
	if [ -z "$SELF_SERVER_ID" ] ; then
		$LOGGER -p $LOG_FACILITY.err -t "$TAG" "Self server id is not set in $CONF/serverid.cfg"
		return 1
	fi

	[ -z "$host" ] && return 0

	if [ -z "$file" ] ; then
		[ -z "$study" ] && return 0
		file="$TMP/$study.$$.rdmp"
		removeDmpFile="true"
		cat <<-EndOfDumpfile > $file
			(0008,0052) CS [STUDY]
			(0020,000d) UI [$study]
		EndOfDumpfile
	fi

	[ ! -f "$file" ] && return 0

	. $DICOM_VAR/pb-scp.cfg
	[ -z "$AE_TITLE" ] && AE_TITLE="$DICOM_AE_TITLE"

	$LOGGER -p $LOG_FACILITY.info -t "$TAG" "Start retrieving $study from $host to $logDest"

	studydir=`$REPOSITORYHANDLER_HOME/bin/locateResource $REPOSITORY_DATA_DICOM $study`

	DEST_AE="$destAE"
	if [ -z "$DEST_AE" ] ; then 
		if [ "$destDev" = "SELF" ] ; then
			DEST_AE="$AE_TITLE"
			if [ -n "$study" ] ; then
				# Retrieving a study, check and create storestate.rec file
				# If storestate.rec file exists and ProcessMode is 'Deleted'
				# then fix ProcessMode for the objects to be registered to the
				# database

				ProcessMode="Process"
				[ -f $CONF/self.rec ] && . $CONF/self.rec

				storestateParams="--study \"$study\" --stagestate=\"Staged\" --processmode=\"$ProcessMode\" --retrieved=\"yes\" --source \"@Archive-$host@$SELF_SERVER_ID\" --archive \"@Archive-$host@$SELF_SERVER_ID\""

				if [ ! -f "$studydir/.info/storestate.rec" ] ; then
					# Create new storestate.rec file
					storestateParams="${storestateParams} -C --owner=\"yes\""
				fi

				# Check if the peer device is an External Archive and add it to forward devices 
				Archive="no"
				[ -f "$ARCHIVE_VAR/archive.cfg" ] && . $ARCHIVE_VAR/archive.cfg
				if [ "$Archive" = "external" -a "$ExternalRetrieve" = "$host" ] ; then
					storestateParams="${storestateParams} --fwddevice \"$ExternalStore\""
				fi

				eval $UTILS_HOME/bin/storestate ${storestateParams}

				if [ -n "$studydir" ] ; then
					Date=`date +"%Y%m%d%H%M%S"`
					# Reindex existing objects if any
					for dcmfile in `ls $studydir` ; do
						$UTILS_HOME/bin/storeDICOMFile -e -f "${dcmfile}" -s "${study}" -I "1" -D "DELTA_STORESCP_REG_QR_GET"
					done
				else
					# the study directory did not exist before but storestate could create it
					studydir=`$REPOSITORYHANDLER_HOME/bin/locateResource $REPOSITORY_DATA_DICOM $study`
				fi
			fi # -n "$study"

			# Adding this reference is not really required as it has already been added in ArchiveRetrieve.class
			$UTILS_HOME/bin/refcount -n "archive_${host}@${SELF_SERVER_ID}_get" -t "STUDYDIR" -i "$study" >>$LOG/refcount.error 2>&1
		else
			#destDev != SELF
			DEST_AE=`echo "select AE from $TARGETTBL where ID='$destDev';" |
			$MYSQL_HOME/bin/mysql -N $DBNAME`
		fi
	fi

	if [ -z "$DEST_AE" ] ; then
		log_ "Start retrieve to $logDest failed: Destination AE title is empty"
		return 2
	fi

	DATE=`date +"%Y%m%d%H%M%S"`

	dcmfile=$TMP/move.$DATE.$$.dcm

	$DCMTK_HOME/bin/dump2dcm $file $dcmfile >/dev/null
	res=$?

	[ "$removeDmpFile" = "true" ] && rm $file 2>/dev/null

	if [ $res -ne 0 -o ! -f "$dcmfile" ] ; then
		log_ "Start retrieve to $logDest failed: Error creating query dcm file"
		return 1
	fi

	# Create .info/user file: Dcreg will pick up userid for STUDY level retrieve log entries from it
	[ -n "$user" -a -n "$studydir" ] && echo "$user" > $studydir/.info/user
	
	echo "select HOST, PORT, AE, PROTOCOL from $TARGETTBL where ID='$host';" |
	$MYSQL_HOME/bin/mysql -N $DBNAME |
	if read fhost fport fae fprotocol; then
		
		generalCfgFile=$DICOM_VAR/pb-scp.cfg
		[ -f $generalCfgFile ] && . $generalCfgFile
			
		cfgFile=$DICOM_VAR/devices/$host.cfg
		[ -f $cfgFile ] && . $cfgFile
		RECV_OPT=" -rt $RECEIVE_TIMEOUT"
		SEND_OPT=" -st $SEND_TIMEOUT"
		[ -n "$CONNECTION_TIMEOUT" ] && CONN_OPT=" -ct $CONNECTION_TIMEOUT"
		[ -n "$WAIT_TIMEOUT" ] && WAIT_OPT=" -wt $WAIT_TIMEOUT"

		calledAE=""
		[ -n "$fae" -a "*" != "$fae" ] && calledAE=" --call $fae"

		callingAE="-aet $AE_TITLE"

		TLS_OPT=""
		[ "$fprotocol" = "S" ] && TLS_OPT="+tls $SSL_SITE_KEY $SSL_SITE_CERT -ic +ps"

		if [ "$fprotocol" = "S" ] ; then
			$OPENSSL_PASS
		fi | \
		$DICOM_HOME/bin/movescu -v -S $TLS_OPT $WAIT_OPT --move $DEST_AE $callingAE $calledAE $fhost $fport $dcmfile > $TMP/movescu.res.$$ 2>&1
		ret=$?

		rm $dcmfile

		if [ $ret -ne 0 ] ; then
			log_ "Start retrieve to $logDest failed: Movescu returned $ret"
			return $ret
		fi

		# Check if number of completed sub operations is greater than 0
		# Using extended regexp to ignore optional leading tag ('I:', 'Error:', ...) and/or optional leading spaces
		numCompleted=`sed -rn 's;^([[:alnum:]]+:){0,1} *Completed Suboperations *: *\<([0-9]*)\> *$;\2;p' $TMP/movescu.res.$$  | tail -n 1`
		dimseStatus=`sed -rn 's;^([[:alnum:]]+:){0,1} *DIMSE Status *: *(.*)$;\2;p' $TMP/movescu.res.$$ | tail -n 1`
		strSuccess=`grep -o "Success" <<<"$dimseStatus"`
		strFailed=`grep -o "Unable to process" <<<"$dimseStatus"`

		[[ ! "$numCompleted" =~ ^[0-9]+$ ]] && numCompleted=0

		$LOGGER -p $LOG_FACILITY.debug -t "$TAG" "NumberOfCompletedSubOperations: $numCompleted [$dimseStatus] [$strSuccess] [$strFailed]"
		rm  $TMP/movescu.res.$$
		if [ "$numCompleted" -gt 0 ] && [ -z "$strFailed" ] ; then
			$LOGGER -p $LOG_FACILITY.debug -t "$TAG" "log success: [ -n \"$numCompleted\" -a \"$numCompleted\" -gt 0 -a -z \"$strFailed\" ]"
			log_
			return 0;
		elif [ -n "$strSuccess" ] ; then
			$LOGGER -p $LOG_FACILITY.debug -t "$TAG" "log success: [ -n \"$strSuccess\" ]"
			log_
			return 0;
		fi

		log_ "Start retrieve to $logDest failed: No completed sub operations ($strFailed)"
		return 4

	else
		log_ "Start retrieve to $logDest failed: Error getting device info from database for $host"
		rm $dcmfile
		return 3
	fi

} # run_()

print_()
{
	printFile=""
	[ -n "$file" ] && printFile=" [QR file: $file]"

	if [ "$destDev" = "SELF" ] ; then
		echo "Retrieve study [${study}] from ${host}${printFile}"
	else
		echo "Move study [${study}] from ${host} to ${logDest}${printFile}"
	fi
}

#
# default get resources 
#
get_resources_()
{
	echo "System.NET.Devices:$host"
}


#
# default get display name (for subjob)
#
get_display_name_()
{
	echo "Retrieve/Move study from peer"
}

#
# get property
#
get_property_()
{
	if [ "$PROPERTY" = "1PST" -a -n "$study" ] ; then
		echo -n "$study" 
	elif [ "$PROPERTY" = "1PPE" -a -n "$host" ] ; then
		echo -n "$host"
	else 
		get_def_property_
	fi
}

get_watson_message_()
{
	[ -z "$DR_WATSON_ID" ] && DR_WATSON_ID="move.$host.$logDest"
	[ -z "$DR_WATSON_MESSAGE" ] && DR_WATSON_MESSAGE="Moving studies from $host to $logDest"

	if [ "$RUN_RESULT" -gt 124 ] ; then
		WATSON_MESSAGE_ID="$DR_WATSON_ID.failed"
		WATSON_MESSAGE_TEXT="FAILED: $DR_WATSON_MESSAGE"
		return
	fi

	if [ "$RUN_RESULT" -ne 0 -a "$RETRY_NUM" -gt 0 -a $(( $RETRY_NUM % 10)) -eq 0 ] ; then
		WATSON_MESSAGE_TEXT="ERROR: $DR_WATSON_MESSAGE [Retry #$RETRY_NUM / Result: $RUN_RESULT]"
		WATSON_MESSAGE_ID="$DR_WATSON_ID.$RETRY_NUM"
	fi
}

#
# Override calculate retry min method to stop with OK after the 
# 5th retry even if the 5th retry is failed
#
calculate_retry_min_()
{
	[ -z "$RETRY_NUM" ] && RETRY_NUM=0

	# Return two min for the first 5 retry
	[ $RETRY_NUM -lt 5  ] && return 2    #  2 min

	# Stop retrying after the 5th retry and pretend everything was OK.
	# No cleanup is needed here unless a cleanup_ method is introduced
	# refcount and directories are freed up in run_
	remove_refcount_
	
	return 0
}
